"""Randomization maintenance"""
from contextlib import contextmanager
from functools import wraps

import numpy as np


@contextmanager
def seed(seed_: int = 0):
    """Context manager that creates a seeded scope."""
    state = np.random.get_state()
    np.random.seed(seed_)
    yield
    np.random.set_state(state)


def seeded(wrapped_requires_seed: bool = False):
    """Create seeded scope for function call.

    Parameters
    ----------
    wrapped_requires_seed: bool, optional, default: False
        if true, passes seed parameter to the inner function
    """
    get = dict.get if wrapped_requires_seed else dict.pop

    def _seeded_maker(func):
        @wraps(func)
        def _seeded(*args, **kwargs):
            _seed = get(kwargs, "seed", 0)
            with seed(_seed):
                return func(*args, **kwargs)

        return _seeded

    return _seeded_maker



"""Sampling methods for statistical indices computation purposes"""

from ._core import BaseSampler, ParallelSampler
from ._stratified_sampler import StratifiedSampler
from ._uniform_sampler import UniformPCASampler, UniformSampler

__all__ = [
    "BaseSampler",
    "ParallelSampler",
    "UniformSampler",
    "UniformPCASampler",
    "StratifiedSampler",
]



from functools import partial
from multiprocessing import Pool
from typing import Callable, Tuple

import numpy as np
from tqdm import tqdm

from divik.feature_selection._exims._structness import structness


class pipe:
    def __init__(self, *functions):
        self.functions = functions

    def __call__(self, *args, **kwargs):
        result = self.functions[0](*args, **kwargs)
        for func in self.functions[1:]:
            result = func(result)
        return result


def progress_bar(description: str = None):
    return partial(tqdm, desc=description)


def pmap(func, collection, **kwargs):
    with Pool() as pool:
        return pool.map(func, collection, **kwargs)


def apply(func, collection):
    return [func(element) for element in collection]


def for_each(func, lazy: bool = True, parallel: bool = False, **kwargs):
    if parallel:
        return partial(pmap, func, **kwargs)
    if lazy:
        return partial(map, func)
    else:
        return partial(apply, func)


def as_image(data: np.ndarray, x: np.ndarray, y: np.ndarray, default=-1) -> np.ndarray:
    x, y = x.astype(int), y.astype(int)
    translated_x, translated_y = x - np.min(x), y - np.min(y)
    rows, columns = int(np.max(translated_y) + 1), int(np.max(translated_x) + 1)
    if len(data.shape) < 2:
        data = data.reshape((data.shape[0], 1))
    cube = default * np.ones((rows, columns, data.shape[1]))
    cube[translated_y, translated_x] = data
    return cube


_IGNORED = -1
_Feature = np.ndarray
_Structness = float
_FeatureProcessor = Callable[[_Feature], Tuple[_Structness, _Structness]]
_remove_channel_dimension = partial(np.squeeze, axis=2)


def _feature_processor(x: np.ndarray, y: np.ndarray) -> _FeatureProcessor:
    # noinspection PyTypeChecker
    return pipe(
        partial(as_image, x=x, y=y, default=_IGNORED),
        _remove_channel_dimension,
        partial(structness, ignored=[_IGNORED]),
    )


def _normalize_columns(matrix) -> np.ndarray:
    matrix = np.array(matrix, dtype=float)
    matrix += np.finfo(float).eps
    assert len(matrix.shape) == 2
    return matrix / np.max(matrix, axis=0)


_as_features = np.transpose
_normalize_structness_by_kind = _normalize_columns
_sumarize_structness_by_feature = pipe(partial(np.sum, axis=1), np.ravel)
FeaturesStructness = np.ndarray
_StructnessEstimator = Callable[[np.ndarray], FeaturesStructness]


def _estimator(structness_: _FeatureProcessor) -> _StructnessEstimator:
    # noinspection PyTypeChecker
    return pipe(
        _as_features,
        progress_bar("feature structness"),
        for_each(structness_, parallel=True),
        _normalize_structness_by_kind,
        _sumarize_structness_by_feature,
    )


def exims(data: np.ndarray, x: np.ndarray, y: np.ndarray) -> FeaturesStructness:
    structness_estimator = _estimator(_feature_processor(x, y))
    return structness_estimator(data)



import logging
import os
from functools import partial

import h5py
import numpy as np
import pandas as pd
from scipy import io as scio

import divik.core as u


def _load_mat_with(path: str, backend=scio.loadmat, ignore="__") -> np.ndarray:
    data = backend(path)
    logging.debug("Data file opened successfully.")
    key = [key for key in list(data.keys()) if not key.startswith(ignore)]
    logging.debug("Found variables: {0}".format(key))
    if len(key) != 1:
        raise ValueError(
            "There should be a single variable inside MAT-file: "
            + path
            + "\nWere: "
            + str(key)
        )
    logging.debug("Selecting variable: {0}".format(key[0]))
    selected = data[key[0]]
    logging.debug("Loaded variable from file.")
    contignuous = np.array(selected, dtype=float)
    logging.debug("Converted to contignuous.")
    return contignuous


def _load_mat(path: str) -> np.ndarray:
    logging.debug("Loading MAT-file: " + path)
    try:
        logging.debug("Trying out legacy MAT-file loader.")
        return _load_mat_with(path, backend=scio.loadmat, ignore="__")
    except NotImplementedError:  # v7.3 MATLAB HDF5 MAT-File
        logging.debug("Legacy MAT-file loader failed, restarting with HDF5 loader.")
        hdf5 = partial(h5py.File, mode="r")
        return _load_mat_with(path, backend=hdf5, ignore="#").T


def load_data(path: str) -> u.Data:
    """Load 2D tabular data from file"""
    logging.info("Loading data: " + path)
    normalized = path.lower()
    if normalized.endswith(".csv"):
        loader = partial(np.loadtxt, delimiter=",")
    elif normalized.endswith(".txt"):
        loader = np.loadtxt
    elif normalized.endswith(".npy"):
        loader = np.load
    elif normalized.endswith(".mat"):
        loader = _load_mat
    else:
        message = "Unsupported data format: " + os.path.splitext(path)[1]
        logging.error(message)
        raise IOError(message)
    return loader(path)


def try_load_data(path):
    """Load 2D tabular data from file with logging"""
    try:
        data = load_data(path)
        logging.debug("Data loaded successfully.")
    except Exception as ex:
        logging.error("Data loading failed with an exception.")
        logging.error(repr(ex))
        raise
    return data


def try_load_xy(path):
    """Load integer spatial coordinates with logging from file"""
    if path is not None:
        try:
            xy = load_data(path).astype(int)
            logging.debug("Coordinates loaded successfully.")
        except Exception as ex:
            logging.error("Coordinates loading failed with an exception.")
            logging.error(repr(ex))
            raise
    else:
        xy = None
    return xy


def save_csv(array: np.ndarray, fname: str):
    """Save array to csv"""
    pd.DataFrame(array).to_csv(fname, header=False, index=False)



import numpy as np
import pandas as pd
from sklearn.base import (
    BaseEstimator,
    ClusterMixin,
    clone,
)

from divik.core import Subsets, configurable

_DEFAULT = object()


def _get_first_attr(obj, prop_name_candidates, default=_DEFAULT):
    for prop_name in prop_name_candidates:
        try:
            return getattr(obj, prop_name)
        except AttributeError:
            pass  # purposeful silence
    if default is _DEFAULT:
        raise AttributeError(
            f"{prop_name_candidates} do not exist in {obj.__class__.__name__}"
        )
    return default


def _get_final_estimator(estimator):
    try:  # sklearn.Pipeline or similar
        return estimator[-1]
    except TypeError:
        return estimator


@configurable
class TwoStep(BaseEstimator, ClusterMixin):
    """Perform a two-step clustering with a given clusterer

    Separates a dataset into ``n_subsets``, processes each of them separately
    and then combines the results.
    
    Works with centroid-based clustering methods, as it requires cluster
    representatives to combine the result.

    Parameters
    ----------
    clusterer : Union[AutoKMeans, Pipeline, KMeans]
        A centroid-based estimator for the purpose of clustering.

    n_subsets : int, default 10
        The number of subsets into which the original dataset should be
        separated
    
    random_state : int, default 42
        Random state to use for seeding the random number generator.
    
    Examples
    --------
    >>> from sklearn.datasets import make_blobs
    >>> from divik.cluster import KMeans, TwoStep
    >>> X, _ = make_blobs(
    ...     n_samples=10_000, n_features=2, centers=3, random_state=42
    ... )
    >>> kmeans = KMeans(n_clusters=3)
    >>> ctr = TwoStep(kmeans).fit(X)
    """
    def __init__(self, clusterer, n_subsets: int = 10, random_state: int = 42):
        self.clusterer = clusterer
        self.n_subsets = n_subsets
        self.random_state = random_state

    def _label_in_subsets(self, X):
        subsets = Subsets(n_splits=self.n_subsets, random_state=self.random_state)
        X_sct = subsets.scatter(X)

        labels_part = [clone(self.clusterer).fit_predict(X_) for X_ in X_sct]
        sct_groups = [l * 0 + i for i, l in enumerate(labels_part)]

        labels = subsets.combine(labels_part)
        groups = subsets.combine(sct_groups)

        cross_fold_labels = [f"g{g}_l{l}" for l, g in zip(labels, groups)]
        return cross_fold_labels

    def fit(self, X, y=None):
        initial_labels = self._label_in_subsets(X)
        centroids = pd.DataFrame(X).groupby(initial_labels).mean()
        self.estimator_ = clone(self.clusterer)
        centroids_labels = self.estimator_.fit_predict(centroids)
        to_final = dict(zip(centroids.index, centroids_labels))
        final_labels = np.array([to_final[l] for l in initial_labels])
        self.labels_ = final_labels
        self.n_clusters_ = _get_first_attr(
            _get_final_estimator(self.estimator_),
            ['n_clusters', 'n_clusters_'],
        )
        return self

    def predict(self, X, y=None):
        return self.estimator_.predict(X)

    def fit_predict(self, X, y=None):
        return self.fit(X, y).labels_



"""multiprocessing.Pool helpers"""
import ctypes
import os
import uuid
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager

# RawArray exists, but PyCharm goes crazy
# noinspection PyUnresolvedReferences
from multiprocessing import Pool, RawArray

import numpy as np


class SharedArrayWrapper(metaclass=ABCMeta):
    @property
    @abstractmethod
    def value(self):
        """Get data (can be called by the recipient)

        Recipient must not modify the data.
        """
        pass


class SharedArray(metaclass=ABCMeta):
    @abstractmethod
    def store(self, data: np.ndarray) -> SharedArrayWrapper:
        """Store the data (must be called by the owner)"""
        pass

    @abstractmethod
    def purge(self):
        """Purge data (must be called by the owner)"""
        pass


class PosixSharedArrayWrapper(SharedArrayWrapper):
    def __init__(self, ref: str):
        self._ref = ref

    @property
    def value(self):
        return PosixSharedArray.DATA[self._ref]


class PosixSharedArray(SharedArray):
    DATA = {}

    def __init__(self):
        self._ref = str(uuid.uuid4())

    def store(self, data):
        if self._ref in PosixSharedArray.DATA:
            raise RuntimeError("UnixSharedArray already stores value")
        PosixSharedArray.DATA[self._ref] = data
        return PosixSharedArrayWrapper(self._ref)

    def purge(self):
        if self._ref in PosixSharedArray.DATA:
            del PosixSharedArray.DATA[self._ref]


class WinSharedArrayWrapper(SharedArrayWrapper):
    def __init__(self, shape, dtype, array):
        self._shape = shape
        self._dtype = dtype
        self._array = array

    @property
    def value(self):
        return np.frombuffer(self._array, dtype=self._dtype).reshape(self._shape)


class WinSharedArray(SharedArray):
    _DTYPES = {
        np.dtype(np.float64): ctypes.c_double,
        np.dtype(np.float32): ctypes.c_float,
        np.dtype(np.uint64): ctypes.c_uint64,
        np.dtype(np.uint32): ctypes.c_uint32,
        np.dtype(np.uint16): ctypes.c_uint16,
        np.dtype(np.int64): ctypes.c_int64,
        np.dtype(np.int32): ctypes.c_int32,
        np.dtype(np.int16): ctypes.c_int16,
    }

    def __init__(self):
        self._shape = None
        self._dtype = None
        self._array = None

    def store(self, data):
        if self._array is not None:
            raise RuntimeError("WindowsSharedArray already stores value")
        if data.dtype not in WinSharedArray._DTYPES:
            raise ValueError("Unsupported data type")
        self._shape = data.shape
        self._dtype = data.dtype
        ctype = WinSharedArray._DTYPES[data.dtype]
        self._array = RawArray(ctype, data.size)
        numpied = np.frombuffer(self._array, dtype=self._dtype)
        np.copyto(numpied, data.ravel())
        return WinSharedArrayWrapper(self._shape, self._dtype, self._array)

    def purge(self):
        self._shape = None
        self._dtype = None
        self._array = None


def _make_shared_array() -> SharedArray:
    if os.name == "posix":
        return PosixSharedArray()
    return WinSharedArray()


@contextmanager
def share(array: np.ndarray):
    """Share a numpy array between ``multiprocessing.Pool`` processes"""
    shared = _make_shared_array()
    try:
        yield shared.store(array)
    except Exception:
        raise
    finally:
        shared.purge()


def get_n_jobs(n_jobs):
    """Determine the actual number of possible jobs"""
    n_cpu = os.cpu_count() or 1
    n_jobs = 1 if n_jobs is None else n_jobs
    if n_jobs <= 0:
        n_jobs = min(n_jobs + 1 + n_cpu, n_cpu)
    n_jobs = n_jobs or n_cpu
    return n_jobs


class DummyPool:
    def __init__(self, processes, initializer=None, initargs=None, *args, **kwargs):
        if initargs is None:
            initargs = ()
        if initializer is not None:
            initializer(*initargs)

    def apply(self, func, args, kwds):
        return func(*args, **kwds)

    # noinspection PyUnusedLocal
    def map(self, func, iterable, chunksize=None):
        return [func(v) for v in iterable]

    # noinspection PyUnusedLocal
    def starmap(self, func, iterable, chunksize=None):
        return [func(*v) for v in iterable]


@contextmanager
def maybe_pool(processes: int = None, *args, **kwargs):
    """Create ``multiprocessing.Pool`` if multiple CPUs are allowed

    Examples
    --------

    >>> from divik.core import maybe_pool
    >>> with maybe_pool(processes=1) as pool:
    ...     # Runs in sequential
    ...     pool.map(id, range(10000))
    >>> with maybe_pool(processes=-1) as pool:
    ...     # Runs with all cores
    ...     pool.map(id, range(10000))
    """
    n_jobs = get_n_jobs(processes)
    if n_jobs == 1 or n_jobs == 0:
        yield DummyPool(n_jobs, *args, **kwargs)
    else:
        with Pool(n_jobs, *args, **kwargs) as pool:
            yield pool



import inspect

import gin
import six


# TODO: Delete this file when https://github.com/google/gin-config/pull/48 gets merged
def _decorate_fn_or_cls(decorator, fn_or_cls, subclass=False):
    """Decorate a function or class with the given decorator.

    When `fn_or_cls` is a function, applies `decorator` to the function and
    returns the (decorated) result.

    When `fn_or_cls` is a class and the `subclass` parameter is `False`, this will
    replace `fn_or_cls.__init__` with the result of applying `decorator` to it.

    When `fn_or_cls` is a class and `subclass` is `True`, this will subclass the
    class, but with `__init__` defined to be the result of applying `decorator` to
    `fn_or_cls.__init__`. The decorated class has metadata (docstring, name, and
    module information) copied over from `fn_or_cls`. The goal is to provide a
    decorated class the behaves as much like the original as possible, without
    modifying it (for example, inspection operations using `isinstance` or
    `issubclass` should behave the same way as on the original class).

    Args:
      decorator: The decorator to use.
      fn_or_cls: The function or class to decorate.
      subclass: Whether to decorate classes by subclassing. This argument is
        ignored if `fn_or_cls` is not a class.

    Returns:
      The decorated function or class.
    """
    if not inspect.isclass(fn_or_cls):  # pytype: disable=wrong-arg-types
        return decorator(gin.config._ensure_wrappability(fn_or_cls))

    construction_fn = gin.config._find_class_construction_fn(fn_or_cls)

    if subclass:
        klass = fn_or_cls
        while hasattr(klass, "_gin_decorated") and klass._gin_decorated:
            klass = klass.__bases__[0]

        class DecoratedClass(fn_or_cls):
            __doc__ = fn_or_cls.__doc__
            __module__ = fn_or_cls.__module__
            __class__ = klass

            def __reduce__(self):
                return super(DecoratedClass, self).__reduce__()

        DecoratedClass.__name__ = fn_or_cls.__name__
        DecoratedClass._gin_decorated = True
        DecoratedClass.__qualname__ = fn_or_cls.__qualname__
        cls = DecoratedClass
    else:
        cls = fn_or_cls

    decorated_fn = decorator(gin.config._ensure_wrappability(construction_fn))
    if construction_fn.__name__ == "__new__":
        decorated_fn = staticmethod(decorated_fn)
    setattr(cls, construction_fn.__name__, decorated_fn)
    return cls


gin.config._decorate_fn_or_cls = _decorate_fn_or_cls



import numpy as np
from sklearn.model_selection import KFold


class Subsets:
    """Scatter dataset to disjoint random subsets and combine them back

    Parameters
    ----------
    n_splits : int, default 10
        Number of subsets that will be generated.

    random_state : int, default 42
        Random state to use for seeding the random number generator.

    Examples
    --------
    >>> from divik.core import Subsets
    >>> subsets = Subsets(n_splits=10, random_state=42)
    >>> X_list = subsets.scatter(X)
    >>> len(X_list)
    10
    >>> # do some computations on each subset
    >>> y = subsets.combine(y_list)
    """

    def __init__(self, n_splits: int = 10, random_state: int = 42):
        self.n_splits = n_splits
        self.random_state = random_state

    def _fold_indices(self, X):
        kfold = KFold(
            n_splits=self.n_splits, shuffle=True, random_state=self.random_state
        )
        return [idx for _, idx in kfold.split(X)]

    def scatter(self, X):
        self._n_rows = X.shape[0]
        self._idx = self._fold_indices(X)
        return [X[idx] for idx in self._idx]

    def _validate_shape(self, X_list):
        shapes = [X.shape for X in X_list]
        if np.sum([s[0] for s in shapes]) != self._n_rows:
            raise ValueError("Rows number mismatch")

        dims = [len(s) for s in shapes]
        if np.min(dims) != np.max(dims):
            raise ValueError("Inconsistent ndim")

        data_dims = [s[1:] for s in shapes]
        if np.any(np.min(data_dims, axis=0) != np.max(data_dims, axis=0)):
            raise ValueError("Data shape mismatch")

    def combine(self, X_list):
        self._validate_shape(X_list)
        X = np.zeros((self._n_rows, *X_list[0].shape[1:]), dtype=X_list[0].dtype)
        for idx, X_part in zip(self._idx, X_list):
            X[idx] = X_part
        return X



"""Unsupervised high-throughput data analysis methods"""
import os

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

if os.environ.get('READTHEDOCS') == 'True':
    import toml
    dirname = os.path.dirname(__file__)
    with open(os.path.join(dirname, '../pyproject.toml')) as f:
        __version__ = toml.load(f)['tool']['poetry']['version']
else:
    __version__ = importlib_metadata.version(__name__)

from ._summary import plot, reject_split

__all__ = [
    "__version__",
    "plot",
    "reject_split",
]



"""Unsupervised feature extraction methods"""

from ._histogram import HistogramEqualization
from ._pca import KneePCA
from ._spectral import LocallyAdjustedRbfSpectralEmbedding

__all__ = [
    "HistogramEqualization",
    "KneePCA",
    "LocallyAdjustedRbfSpectralEmbedding",
]



import numpy as np
from sklearn.base import BaseEstimator

from divik.core import build, configurable

from ._gmm_selector import GMMSelector
from ._outlier import OutlierSelector
from ._percentage_selector import PercentageSelector
from ._stat_selector_mixin import NoSelector, SelectorMixin


@configurable
class HighAbundanceAndVarianceSelector(BaseEstimator, SelectorMixin):
    """Feature selector that removes low-mean and low-variance features

    Exercises ``GMMSelector`` to filter out the low-abundance noise features
    and select high-variance informative features.

    This feature selection algorithm looks only at the features (X), not the
    desired outputs (y), and can thus be used for unsupervised learning.

    Parameters
    ----------
    use_log: bool, optional, default: False
        Whether to use the logarithm of feature characteristic instead of the
        characteristic itself. This may improve feature filtering performance,
        depending on the distribution of features, however all the
        characteristics (mean, variance) have to be positive for that -
        filtering will fail otherwise. This is useful for specific cases in
        biology where the distribution of data may actually require this option
        for any efficient filtering.

    min_features: int, optional, default: 1
        How many features must be preserved.

    min_features_rate: float, optional, default: 0.0
        Similar to ``min_features`` but relative to the input data features
        number.

    max_components: int, optional, default: 10
        The maximum number of components used in the GMM decomposition.

    Attributes
    ----------
    abundance_selector_: GMMSelector
        Selector used to filter out the noise component.

    variance_selector_: GMMSelector
        Selector used to filter out the non-informative features.

    selected_: array, shape (n_features,)
        Vector of binary selections of the informative features.

    Examples
    --------
    >>> import numpy as np
    >>> import divik.feature_selection as fs
    >>> np.random.seed(42)
    >>> # Data in this case must be carefully crafted
    >>> labels = np.concatenate([30 * [0] + 20 * [1] + 30 * [2] + 40 * [3]])
    >>> data = np.vstack(100 * [labels * 10.])
    >>> data += np.random.randn(*data.shape)
    >>> sub = data[:, :-40]
    >>> sub += 5 * np.random.randn(*sub.shape)
    >>> # Label 0 has low abundance but high variance
    >>> # Label 3 has low variance but high abundance
    >>> # Label 1 and 2 has not-lowest abundance and high variance
    >>> selector = fs.HighAbundanceAndVarianceSelector().fit(data)
    >>> selector.transform(labels.reshape(1,-1))
    array([[1 1 1 1 1 ...2 2 2]])

    """

    def __init__(
        self,
        use_log: bool = False,
        min_features: int = 1,
        min_features_rate: float = 0.0,
        max_components: int = 10,
    ):
        self.use_log = use_log
        self.min_features = min_features
        self.min_features_rate = min_features_rate
        self.max_components = max_components

    def fit(self, X, y=None):
        """Learn data-driven feature thresholds from X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Sample vectors from which to compute feature characteristic.

        y : any
            Ignored. This parameter exists only for compatibility with
            sklearn.pipeline.Pipeline.

        Returns
        -------
        self
        """
        min_features = max(self.min_features, self.min_features_rate * X.shape[1])

        if min_features == X.shape[1]:
            self.selected_ = np.ones((X.shape[1],), dtype=bool)
            return self

        self.abundance_selector_ = GMMSelector(
            "mean",
            use_log=self.use_log,
            n_candidates=1,
            min_features=min_features,
            preserve_high=True,
            max_components=self.max_components,
        ).fit(X)
        filtered = self.abundance_selector_.transform(X)
        self.selected_ = self.abundance_selector_.selected_.copy()

        self.variance_selector_ = GMMSelector(
            "var",
            use_log=self.use_log,
            n_candidates=None,
            min_features=min_features,
            preserve_high=True,
            max_components=self.max_components,
        ).fit(filtered)
        self.selected_[self.selected_] = self.variance_selector_.selected_

        return self

    def _get_support_mask(self):
        """
        Get the boolean mask indicating which features are selected

        Returns
        -------
        support : boolean array of shape [# input features]
            An element is True iff its corresponding feature is selected for
            retention.
        """
        return self.selected_


EPS = 10e-6


# noinspection PyAttributeOutsideInit
@configurable
class OutlierAbundanceAndVarianceSelector(BaseEstimator, SelectorMixin):
    def __init__(
        self, use_log: bool = False, min_features_rate: float = 0.01, p: float = 0.2
    ):
        self.use_log = use_log
        self.min_features_rate = min_features_rate
        self.p = p

    def fit(self, X, y=None):
        """Learn data-driven feature thresholds from X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Sample vectors from which to compute feature characteristic.

        y : any
            Ignored. This parameter exists only for compatibility with
            sklearn.pipeline.Pipeline.

        Returns
        -------
        self
        """
        self.abundance_selector_, a_selected = self._fit_abundance(X)
        filtered = X[:, a_selected]
        self.variance_selector_, v_selected = self._fit_variance(filtered, a_selected)
        self.selected_ = a_selected
        self.selected_[a_selected] = v_selected
        return self

    def _fit_abundance(self, X):
        selector = OutlierSelector(
            stat="mean", use_log=self.use_log, keep_outliers=False
        ).fit(X)
        selected = selector.selected_
        inlier = selector.vals_[selected][0]
        over_inlier = selector.vals_ > inlier
        selected[over_inlier] = True
        p = selected.mean()
        if p < self.min_features_rate or p >= 1 - EPS:
            selector = PercentageSelector(
                stat="mean", use_log=self.use_log, keep_top=True, p=1.0 - self.p
            ).fit(X)
            selected = selector.selected_
        return selector, selected

    def _fit_variance(self, X, old_selected):
        corrected_min = self.min_features_rate / old_selected.mean()
        corrected_p = self.p / old_selected.mean()

        selector = OutlierSelector(
            stat="var", use_log=self.use_log, keep_outliers=True
        ).fit(X)
        selected = selector.selected_
        inlier = selector.vals_[selected == 0][0]
        under_inlier = selector.vals_ < inlier
        selected[under_inlier] = False
        p = selected.mean()

        if p < corrected_min or p >= 1 - EPS:
            selector = PercentageSelector(
                stat="var", use_log=self.use_log, keep_top=True, p=corrected_p
            ).fit(X)
            selected = selector.selected_
        return selector, selected

    def _get_support_mask(self):
        """
        Get the boolean mask indicating which features are selected

        Returns
        -------
        support : boolean array of shape [# input features]
            An element is True iff its corresponding feature is selected for
            retention.
        """
        return self.selected_


def make_specialized_selector(name, n_features, **kwargs):
    """Create a selector by name (``gmm``, ``outlier``, ``none`` or ``auto``)

    ``auto`` switches to ``gmm`` if there is more than 250 features, ``outlier`` below.
    """
    if name == "auto":
        name = "gmm" if n_features > 250 else "outlier"
    filter_cls = {
        "gmm": HighAbundanceAndVarianceSelector,
        "outlier": OutlierAbundanceAndVarianceSelector,
        "none": NoSelector,
    }[name]
    return build(filter_cls, **kwargs)



import logging

import numpy as np
from scipy.spatial import distance as dist
from sklearn.base import BaseEstimator
from sklearn.manifold import SpectralEmbedding
from sklearn.utils.validation import check_is_fitted

from divik.core import configurable
from divik.core.io import save_csv


@configurable
class LocallyAdjustedRbfSpectralEmbedding(BaseEstimator):
    """Spectral embedding for non-linear dimensionality reduction.

    Forms an affinity matrix given by the specified function and
    applies spectral decomposition to the corresponding graph laplacian.
    The resulting transformation is given by the value of the
    eigenvectors for each data point.

    Note : Laplacian Eigenmaps is the actual algorithm implemented here.

    Parameters
    -----------
    distance : {'braycurtis', 'canberra', 'chebyshev', 'cityblock',
    'correlation', 'cosine', 'dice', 'euclidean', 'hamming', 'jaccard',
    'kulsinski', 'mahalanobis', 'atching', 'minkowski', 'rogerstanimoto',
    'russellrao', 'sokalmichener', 'sokalsneath', 'sqeuclidean', 'yule'}
        Distance measure, defaults to ``euclidean``. These are the distances
        supported by scipy package.


    n_components : integer, default: 2
        The dimension of the projected subspace.

    random_state : int, RandomState instance or None, optional, default: None
        A pseudo random number generator used for the initialization of the
        lobpcg eigenvectors.  If int, random_state is the seed used by the
        random number generator; If RandomState instance, random_state is the
        random number generator; If None, the random number generator is the
        RandomState instance used by ``np.random``. Used when ``solver`` ==
        ``amg``.

    eigen_solver : {None, 'arpack', 'lobpcg', or 'amg'}
        The eigenvalue decomposition strategy to use. AMG requires pyamg
        to be installed. It can be faster on very large, sparse problems,
        but may also lead to instabilities.

    n_neighbors : int, default : max(n_samples/10 , 1)
        Number of nearest neighbors for nearest_neighbors graph building.

    n_jobs : int, optional (default = 1)
        The number of parallel jobs to run.
        If ``-1``, then the number of jobs is set to the number of CPU cores.

    Attributes
    ----------

    embedding_ : array, shape = (n_samples, n_components)
        Spectral embedding of the training matrix.

    References
    ----------

    - A Tutorial on Spectral Clustering, 2007
      Ulrike von Luxburg
      http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.165.9323

    - On Spectral Clustering: Analysis and an algorithm, 2001
      Andrew Y. Ng, Michael I. Jordan, Yair Weiss
      http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.19.8100

    - Normalized cuts and image segmentation, 2000
      Jianbo Shi, Jitendra Malik
      http://citeseer.ist.psu.edu/viewdoc/summary?doi=10.1.1.160.2324
    """

    def __init__(
        self,
        distance: str = "euclidean",
        n_components=2,
        random_state=None,
        eigen_solver: str = None,
        n_neighbors: int = None,
        n_jobs: int = 1,
    ):
        self.distance = distance
        self.n_components = n_components
        self.random_state = random_state
        self.eigen_solver = eigen_solver
        self.n_neighbors = n_neighbors
        self.n_jobs = n_jobs

    # noinspection PyAttributeOutsideInit
    def fit(self, X, y=None):
        """Fit the model from data in X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples
            and n_features is the number of features.

        Y: Ignored.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        logging.debug("Computing locally adjusted affinities.")
        d = dist.squareform(dist.pdist(X, metric=self.distance))

        if 0 <= self.n_components <= 1:
            n_components = max(int(self.n_components * X.shape[1]), 1)
        else:
            n_components = self.n_components

        logging.debug("Computing embedding of affinities.")
        embedder = SpectralEmbedding(
            n_components=n_components,
            affinity="precomputed_nearest_neighbors",
            gamma=None,
            random_state=self.random_state,
            eigen_solver=self.eigen_solver,
            n_neighbors=self.n_neighbors,
            n_jobs=self.n_jobs,
        )
        self.embedding_ = embedder.fit_transform(d)
        return self

    def fit_transform(self, X, y=None):
        """Fit the model from data in X and transform X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples
            and n_features is the number of features.

        Y: Ignored.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)
        """
        return self.fit(X).embedding_

    def transform(self, X, y=None):
        if not hasattr(self, "embedding_") or self.embedding_.shape[0] != X.shape[0]:
            self.fit(X, y)
        return self.embedding_

    def save(self, destination: str):
        """Save embedding to a directory

        Parameters
        ----------
        destination : str
            Directory to save the embedding.
        """
        logging.info("Saving embedding to {0}.".format(destination))
        check_is_fitted(self)
        import os
        import pickle
        from functools import partial

        fname = partial(os.path.join, destination)

        logging.debug("Saving model.")
        with open(fname("model.pkl"), "wb") as pkl:
            pickle.dump(self, pkl)

        logging.debug("Saving embedding.")
        save_csv(self.embedding_, fname("embedding.csv"))
        np.save(fname("embedding.npy"), self.embedding_)



"""Unsupervised feature selection methods"""
from ._exims import EximsSelector
from ._gmm_selector import GMMSelector
from ._outlier import OutlierSelector, huberta_outliers
from ._percentage_selector import PercentageSelector
from ._specialized import (
    HighAbundanceAndVarianceSelector,
    OutlierAbundanceAndVarianceSelector,
    make_specialized_selector,
)
from ._stat_selector_mixin import (
    NoSelector,
    SelectorMixin,
    StatSelectorMixin,
)

__all__ = [
    "SelectorMixin",
    "StatSelectorMixin",
    "NoSelector",
    "EximsSelector",
    "GMMSelector",
    "huberta_outliers",
    "OutlierSelector",
    "PercentageSelector",
    "HighAbundanceAndVarianceSelector",
    "OutlierAbundanceAndVarianceSelector",
    "make_specialized_selector",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['divik',
 'divik._cli',
 'divik.cluster',
 'divik.cluster._divik',
 'divik.cluster._kmeans',
 'divik.core',
 'divik.core.io',
 'divik.feature_extraction',
 'divik.feature_selection',
 'divik.feature_selection._exims',
 'divik.sampler',
 'divik.score']

package_data = \
{'': ['*']}

install_requires = \
['dask-distance>=0.2.0,<0.3.0',
 'dask[dataframe]>=2.14.0',
 'h5py>=2.8.0',
 'joblib>=1.0.0,<2.0.0',
 'kneed>=0.5.1',
 'matplotlib>=3.3.3,<4.0.0',
 'numpy>=0.12.1',
 'pandas>=0.20.3',
 'scikit-image>=0.14.1',
 'scikit-learn>=0.19.0',
 'scipy>=0.19.1',
 'tqdm>=4.11.2']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.0,<2.0']}

entry_points = \
{'console_scripts': ['fit-clusters = divik._cli.fit_clusters:main']}

setup_kwargs = {
    'name': 'divik',
    'version': '3.0.1',
    'description': 'Divisive iK-means algorithm implementation',
    'long_description': '[![CodeFactor](https://www.codefactor.io/repository/github/gmrukwa/divik/badge)](https://www.codefactor.io/repository/github/gmrukwa/divik)\n[![BCH compliance](https://bettercodehub.com/edge/badge/gmrukwa/divik?branch=master)](https://bettercodehub.com/)\n[![Maintainability](https://api.codeclimate.com/v1/badges/4cf5d42d0a0076c38445/maintainability)](https://codeclimate.com/github/gmrukwa/divik/maintainability)\n![](https://github.com/gmrukwa/divik/workflows/Build%20and%20push%20deployment%20images/badge.svg)\n![](https://github.com/gmrukwa/divik/workflows/Run%20unit%20tests/badge.svg)\n[![Documentation Status](https://readthedocs.org/projects/divik/badge/?version=latest)](https://divik.readthedocs.io/en/latest/?badge=latest)\n\n# divik\n\nPython implementation of Divisive iK-means (DiviK) algorithm.\n\n## Tools within this package\n\n- Clustering at your command line with fit-clusters\n- Set of algorithm implementations for unsupervised analyses\n  - Clustering\n    - DiviK - hands-free clustering method with built-in feature selection\n    - K-Means with Dunn method for selecting the number of clusters\n    - K-Means with GAP index for selecting the number of clusters\n    - Modular K-Means implementation with custom distance metrics and initializations\n  - Feature extraction\n    - PCA with knee-based components selection\n    - Locally Adjusted RBF Spectral Embedding\n  - Feature selection\n    - EXIMS\n    - Gaussian Mixture Model based data-driven feature selection\n      - High Abundance And Variance Selector - allows you to select highly variant features above noise level, based on GMM-decomposition\n    - Outlier based Selector\n      - Outlier Abundance And Variance Selector - allows you to select highly variant features above noise level, based on outlier detection\n    - Percentage based Selector - allows you to select highly variant features above noise level with your predefined thresholds for each\n  - Sampling\n    - StratifiedSampler - generates samples of fixed number of rows from given dataset\n    - UniformPCASampler - generates samples of random observations within boundaries of an original dataset, and preserving the rotation of the data\n    - UniformSampler - generates samples of random observations within boundaries of an original dataset\n\n## Installation\n\n### Docker\n\nThe recommended way to use this software is through\n[Docker](https://www.docker.com/). This is the most convenient way, if you want\nto use `divik` application.\n\nTo install latest stable version use:\n\n```bash\ndocker pull gmrukwa/divik\n```\n\n### Python package\n\nPrerequisites for installation of base package:\n\n- Python 3.6 / 3.7 / 3.8\n- compiler capable of compiling the native C code and OpenMP support\n\n#### Installation of OpenMP for Ubuntu / Debian\n\nYou should have it already installed with GCC compiler, but if somehow\nnot, try the following:\n\n```bash\nsudo apt-get install libgomp1\n```\n\n#### Installation of OpenMP for Mac\n\nOpenMP is available as part of LLVM. You may need to install it with conda:\n\n```bash\nconda install -c conda-forge "compilers>=1.0.4,!=1.1.0" llvm-openmp\n```\n\n#### DiviK Installation\n\nHaving prerequisites installed, one can install latest base version of the\npackage:\n\n```bash\npip install divik\n```\n\nIf you want to have compatibility with\n[`gin-config`](https://github.com/google/gin-config), you can install\nnecessary extras with:\n\n```bash\npip install divik[gin]\n```\n\n**Note:** Remember about `\\` before `[` and `]` in `zsh` shell.\n\nYou can install all extras with:\n\n```bash\npip install divik[all]\n```\n\n## High-Volume Data Considerations\n\nIf you are using DiviK to run the analysis that could fail to fit RAM of your\ncomputer, consider disabling the default parallelism and switch to\n[dask](https://dask.org/). It\'s easy to achieve through configuration:\n\n- set all parameters named `n_jobs` to `1`;\n- set all parameters named `allow_dask` to `True`.\n\n**Note:** Never set `n_jobs>1` and `allow_dask=True` at the same time, the\ncomputations will freeze due to how `multiprocessing` and `dask` handle\nparallelism.\n\n## Known Issues\n\n### Segmentation Fault\n\nIt can happen if the he `gamred_native` package (part of `divik` package) was\ncompiled with different numpy ABI than scikit-learn. This could happen if you\nused different set of compilers than the developers of the scikit-learn\npackage.\n\nIn such a case, a handler is defined to display the stack trace. If the trace\ncomes from `_matlab_legacy.py`, the most probably this is the issue.\n\nTo resolve the issue, consider following the installation instructions once\nagain. The exact versions get updated to avoid the issue.\n\n## Contributing\n\nContribution guide will be developed soon.\n\nFormat the code with:\n\n```bash\nisort -m 3 --fgw 3 --tc .\nblack -t py36 .\n```\n\n## References\n\nThis software is part of contribution made by [Data Mining Group of Silesian\nUniversity of Technology](http://www.zaed.polsl.pl/), rest of which is\npublished [here](https://github.com/ZAEDPolSl).\n\n- [Mrukwa, G. and Polanska, J., 2020. DiviK: Divisive intelligent K-means for\nhands-free unsupervised clustering in biological big data. *arXiv preprint\narXiv:2009.10706.*][1]\n\n[1]: https://arxiv.org/abs/2009.10706\n',
    'author': 'Grzegorz Mrukwa',
    'author_email': 'g.mrukwa@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gmrukwa/divik',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



import logging

import numpy as np
from sklearn.decomposition import PCA
from sklearn.preprocessing import MinMaxScaler

from divik.core import configurable
from divik.core import seed as seed_
from divik.feature_extraction import KneePCA

from ._core import BaseSampler


@configurable
class UniformSampler(BaseSampler):
    """Samples uniformly from the boundaries of the data

    Parameters
    -----------
    n_rows : int, optional (default None)
        Allows to limit the number of rows in the drawn samples

    n_samples : int, optional (default None)
        Allows to limit the number of samples when iterating

    Attributes
    ----------
    shape_ : (n_rows, n_cols)
        Shape of the drawn samples

    scaler_ : MinMaxScaler
        Scaler ensuring the proper ranges
    """

    def __init__(self, n_rows: int = None, n_samples: int = None):
        self.n_rows = n_rows
        self.n_samples = n_samples

    def fit(self, X, y=None):
        """Fit the model from data in X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples
            and n_features is the number of features.

        Y: Ignored.

        Returns
        -------
        self : UniformSampler
            Returns the instance itself.
        """
        if self.n_rows is None:
            n_rows = X.shape[0]
        else:
            n_rows = self.n_rows
        self.shape_ = n_rows, X.shape[1]
        self.scaler_ = MinMaxScaler().fit(X)
        return self

    def get_sample(self, seed):
        """Return specific sample

        Parameters
        ----------
        seed : int
            The seed to use to draw the sample

        Returns
        -------
        sample : array_like, (*self.shape_)
            Returns the drawn sample
        """
        with seed_(seed):
            unscaled = np.random.random_sample(self.shape_)
        return self.scaler_.inverse_transform(unscaled)


@configurable
class UniformPCASampler(BaseSampler):
    """Rotation-invariant uniform sampling

    Parameters
    -----------
    n_rows : int, optional (default None)
        Allows to limit the number of rows in the drawn samples

    n_samples : int, optional (default None)
        Allows to limit the number of samples when iterating

    whiten : bool, optional (default False)
        When True (False by default) the `pca_.components_` vectors are
        multiplied by the square root of n_samples and then divided by the
        singular values to ensure uncorrelated outputs with unit
        component-wise variances.

        Whitening will remove some information from the transformed signal
        (the relative variance scales of the components) but can sometime
        improve the predictive accuracy of the downstream estimators by
        making their data respect some hard-wired assumptions.

    refit : bool, optional (default False)
        When True (False by default) the `pca_` is re-fit with the smaller
        number of components. This could reduce memory footprint, but
        requires training fitting PCA.

    pca: {'knee', 'full'}, default 'knee'
        Specifies whether to train full or knee PCA.

    Attributes
    ----------
    pca_ : KneePCA or PCA
        PCA transform which provided rotation-invariance

    sampler_ : UniformSampler
        Sampler from the transformed distribution
    """

    def __init__(
        self,
        n_rows: int = None,
        n_samples: int = None,
        whiten: bool = False,
        refit: bool = False,
        pca: str = "knee",
    ):
        self.n_rows = n_rows
        self.n_samples = n_samples
        self.whiten = whiten
        self.refit = refit
        self.pca = pca

    def _make_pca(self):
        if self.pca == "knee":
            return KneePCA(whiten=self.whiten, refit=self.refit)
        if self.pca == "full":
            # Note: random_state is not used in this config!
            return PCA(
                n_components=None,
                copy=True,
                whiten=self.whiten,
                svd_solver="full",
                tol=0.0,
                iterated_power="auto",
                random_state=None,
            )
        logging.error("Unsupported pca value: {}".format(self.pca))
        raise ValueError("Unsupported pca value: {}".format(self.pca))

    def fit(self, X, y=None):
        """Fit the model from data in X.

        PCA is fit to estimate the rotation and UniformSampler is
        fit to transformed data.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples
            and n_features is the number of features.

        Y: Ignored.

        Returns
        -------
        self : UniformPCASampler
            Returns the instance itself.
        """
        self.pca_ = self._make_pca()
        transformed = self.pca_.fit_transform(X)
        self.sampler_ = UniformSampler(n_rows=self.n_rows).fit(transformed)
        return self

    def get_sample(self, seed):
        """Return specific sample

        Sample is generated from transformed distribution and transformed
        back to the original space.

        Parameters
        ----------
        seed : int
            The seed to use to draw the sample

        Returns
        -------
        sample : array_like, (*self.shape_)
            Returns the drawn sample
        """
        transformed = self.sampler_.get_sample(seed)
        return self.pca_.inverse_transform(transformed)



import logging
from functools import partial
from typing import Union

import numpy as np
import pandas as pd
from scipy.spatial import distance as dist

from divik.core import (
    Data,
    configurable,
    maybe_pool,
)
from divik.sampler import BaseSampler, StratifiedSampler

KMeans = "divik.cluster.KMeans"
_BIG_PRIME = 49277


def _inter_centroid(kmeans: KMeans, data: Data, labels=None):
    d = dist.pdist(kmeans.cluster_centers_, kmeans.distance)
    return np.min(d[d != 0])


def _inter_closest(kmeans: KMeans, data: Data, labels=None):
    if labels is None:
        labels = kmeans.labels_
    d = np.inf
    for label in np.arange(kmeans.n_clusters - 1):
        grp = label == labels
        non_grp = label < labels
        dst = dist.cdist(data[grp], data[non_grp], metric=kmeans.distance)
        d = np.minimum(d, dst.min())
    return d


def _intra_avg(kmeans: KMeans, data: Data, labels=None):
    if labels is None:
        labels = kmeans.labels_
    clusters = pd.DataFrame(data).groupby(labels).apply(np.asarray)
    return np.max(
        [
            np.mean(dist.cdist(cluster, centroid.reshape(1, -1), kmeans.distance))
            for cluster, centroid in zip(clusters, kmeans.cluster_centers_)
        ]
    )


def _intra_furthest(kmeans: KMeans, data: Data, labels=None):
    def max_distance(group):
        group = np.asarray(group)
        d = dist.pdist(group, metric=kmeans.distance)
        # 0 is intracluster distance for cluster with one observation
        return np.max(d, initial=0.0)

    if labels is None:
        labels = kmeans.labels_
    return pd.DataFrame(data).groupby(labels).apply(max_distance).max()


_INTER = {
    "centroid": _inter_centroid,
    "closest": _inter_closest,
}
_INTRA = {
    "avg": _intra_avg,
    "furthest": _intra_furthest,
}


@configurable
def dunn(kmeans: KMeans, data: Data, inter="centroid", intra="avg") -> float:
    """Compute Dunn's index for the clustering

    Parameters
    ----------
    kmeans : KMeans
        KMeans object fitted to the data

    data : array, shape (n_samples, n_features)
        Clustered data

    inter : {'centroid', 'closest'}
        Method of computing intercluster distance
        - centroid - uses distances between centroids
        - closest - uses distance between closest members of separate clusters

    intra : {'avg', 'furthest}
        Method of computing intracluster distance
        - avg - uses average distance to the centroid
        - furthest - uses distance between the furthest cluster members

    Returns
    -------
    dunn_index : float
        Value of Dunn's index for the clustering of data
    """
    if kmeans.cluster_centers_.shape[0] == 1:
        return -np.inf
    if inter not in _INTER:
        msg = (
            f"Unsupported intercluster distance {inter}. "
            + f"Supported: {list(_INTER.keys())}"
        )
        logging.error(msg)
        raise ValueError(msg)
    if intra not in _INTRA:
        msg = (
            f"Unsupported intracluster distance {intra}. "
            + f"Supported: {list(_INTRA.keys())}"
        )
        logging.error(msg)
        raise ValueError(msg)
    intercluster = _INTER[inter](kmeans, data)
    intracluster = _INTRA[intra](kmeans, data)
    score = intercluster / intracluster
    return score


def _sample_distances(
    seed: int, sampler: BaseSampler, kmeans: KMeans, inter="centroid", intra="avg"
):
    data = sampler.get_sample(seed)
    labels = kmeans.predict(data)
    inter_ = _INTER[inter](kmeans, data, labels)
    intra_ = _INTRA[intra](kmeans, data, labels)
    return inter_, intra_


@configurable
def sampled_dunn(
    kmeans: KMeans,
    data: Data,
    sample_size: Union[int, float] = 1000,
    n_jobs: int = None,
    seed: int = 0,
    n_trials: int = 10,
    inter="closest",
    intra="furthest",
) -> float:
    data_ = StratifiedSampler(n_rows=sample_size, n_samples=n_trials).fit(
        data, kmeans.labels_
    )
    seeds = list(seed + np.arange(n_trials) * _BIG_PRIME)
    with data_.parallel() as d, maybe_pool(
        n_jobs, initializer=d.initializer, initargs=d.initargs
    ) as pool:
        distances = partial(
            _sample_distances, sampler=d, kmeans=kmeans, inter=inter, intra=intra
        )
        inter_, intra_ = np.array(pool.map(distances, seeds)).T
    s_inter = inter_.std()
    s_intra = intra_.std()
    return (inter_.min() - s_inter) / (intra_.max() + s_intra)



import inspect
import logging
import os
import sys
import time
from contextlib import contextmanager

# RawArray exists, but PyCharm goes crazy
# noinspection PyUnresolvedReferences
from multiprocessing import Pool, RawArray

import numpy as np
import tqdm
from skimage.color import label2rgb

from .. import __version__
from ._types import Data


def normalize_rows(data: Data) -> Data:
    """Translate and scale rows to zero mean and vector length equal one"""
    normalized = data - data.mean(axis=1)[:, np.newaxis]
    norms = np.sum(np.abs(normalized) ** 2, axis=-1, keepdims=True) ** (1.0 / 2)
    normalized /= norms
    return normalized


def visualize(label, xy, shape=None):
    """Create RGB map of labels over with given coordinates"""
    x, y = xy.T
    if shape is None:
        shape = np.max(y) + 1, np.max(x) + 1
    y = y.max() - y
    label = label - label.min() + 1
    label_map = np.zeros(shape, dtype=int)
    label_map[y, x] = label
    image = label2rgb(label_map, bg_label=0)
    return image


@contextmanager
def context_if(condition, context, *args, **kwargs):
    """Create context with given params only if the condition is ``True``"""
    if condition:
        with context(*args, **kwargs) as c:
            yield c
    else:
        yield None


def build(klass, **kwargs):
    """Build instance of klass using matching kwargs"""
    known_param_names = list(inspect.signature(klass).parameters.keys())
    known_params = {n: kwargs[n] for n in known_param_names if n in kwargs}
    return klass(**known_params)


def prepare_destination(
    destination: str, omit_datetime: bool = False, exist_ok: bool = False
) -> str:
    if not omit_datetime:
        datetime = time.strftime("%Y%m%d-%H%M%S")
        destination = os.path.join(destination, datetime)
    os.makedirs(destination, exist_ok=exist_ok)
    return destination


_PRECISE_FORMAT = (
    "%(asctime)s [%(levelname)s] %(filename)40s:%(lineno)3s"
    + " - %(funcName)40s\t%(message)s"
)
_INFO_FORMAT = "%(asctime)s [%(levelname)s]\t%(message)s"


def _file_handler(destination: str):
    log_destination = os.path.join(destination, "logs.txt")
    handler = logging.FileHandler(filename=log_destination, mode="a")
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter(_PRECISE_FORMAT))
    return handler


def _stream_handler(verbose: bool):
    handler = logging.StreamHandler(tqdm.tqdm)
    if verbose:
        handler.setLevel(logging.INFO)
        handler.setFormatter(logging.Formatter(_INFO_FORMAT))
    else:
        handler.setLevel(logging.CRITICAL)
        handler.setFormatter(logging.Formatter(_INFO_FORMAT))
    return handler


def setup_logger(destination: str, verbose: bool = False):
    stream_handler = _stream_handler(verbose)
    file_handler = _file_handler(destination)
    handlers = [
        stream_handler,
        file_handler,
    ]
    del logging.root.handlers[:]
    logging.basicConfig(level=logging.DEBUG, handlers=handlers)
    version_notice = "Using " + sys.argv[0] + " (divik, version " + __version__ + ")"
    logging.info(version_notice)



import json
import logging
import os
import pickle
from functools import partial

import numpy as np
import pandas as pd
from sklearn.pipeline import Pipeline

from divik.core import configurable, visualize

_SAVERS = set()


def saver(fn):
    """Register the function as handler for saving model and related summaries

    The saver function should be reusable for different models exhibiting the
    required variables. Rather prefer checking the required attributes than the
    model class.

    Examples
    --------

    >>> from divik.core.io import saver
    >>> @saver
    ... def my_saver(model, destination, **kwargs):
    ...     if not hasattr(model, 'my_custom_field_'):
    ...         return
    ...     if not 'my_param' in kwargs:
    ...         return
    ...     # custom saving logic comes here

    You can also make this function configurable:

    >>> import gin
    >>> from divik.core.io import saver
    >>> @saver
    ... @gin.configurable(allowlist=['my_param'])
    ... def configurable_saver(model, destination, my_param=None, **kwargs):
    ...     if not hasattr(model, 'my_custom_field_'):
    ...         return
    ...     if my_param is None:
    ...         return
    ...     # custom saving logic comes here
    """
    _SAVERS.add(fn)


def save(model, destination, **kwargs):
    """Save model and related summaries into specified destination directory"""
    fname_fn = partial(os.path.join, destination)
    for save_fn in _SAVERS:
        save_fn(model, fname_fn, **kwargs)


@saver
@configurable(allowlist=["enabled"])
def save_pickle(model, fname_fn, enabled=True, **kwargs):
    if not enabled:
        return
    logging.info("Saving model pickle.")
    with open(fname_fn("model.pkl"), "wb") as pkl:
        pickle.dump(model, pkl)


@saver
def save_summary(model, fname_fn, **kwargs):
    if not hasattr(model, "labels_"):
        return
    logging.info("Saving JSON summary.")
    n_clusters = getattr(model, "n_clusters_", np.unique(model.label_).size)
    with open(fname_fn("summary.json"), "w") as smr:
        json.dump(
            {
                "depth": getattr(model, "depth_", 1),
                "number_of_clusters": n_clusters,
                "mean_cluster_size": model.labels_.size / float(n_clusters),
            },
            smr,
        )


@saver
def save_labels(model, fname_fn, **kwargs):
    if not hasattr(model, "labels_"):
        return
    logging.info("Saving final partition.")
    np.save(fname_fn("final_partition.npy"), model.labels_)
    np.savetxt(fname_fn("final_partition.csv"), model.labels_, delimiter=", ", fmt="%i")
    if "xy" in kwargs:
        import skimage.io

        visualization = visualize(model.labels_, xy=kwargs["xy"])
        skimage.io.imsave(fname_fn("final_partition.png"), visualization)


@saver
def save_multiple_labels(model, fname_fn, **kwargs):
    if not hasattr(model, "estimators_") or not hasattr(model.estimators[0], "labels_"):
        return
    logging.info("Saving all considered partitions.")
    part = np.hstack([e.labels_.reshape(-1, 1) for e in model.estimators_])
    np.save(fname_fn("partitions.npy"), part)
    np.savetxt(fname_fn("partitions.csv"), part, delimiter=", ", fmt="%i")

    import skimage.io

    for i in range(part.shape[1]):
        np.savetxt(
            fname_fn("partitions.{0}.csv").format(i),
            part[:, i].reshape(-1, 1),
            delimiter=", ",
            fmt="%i",
        )
        if "xy" in kwargs:
            visualization = visualize(part, xy=kwargs["xy"])
            skimage.io.imsave(fname_fn("partitions.{0}.png").format(i), visualization)


@saver
def save_centroids(model, fname_fn, **kwargs):
    if not hasattr(model, "centroids_"):
        return
    logging.info("Saving centroids.")
    np.save(fname_fn("centroids.npy"), model.centroids_)
    np.savetxt(fname_fn("centroids.csv"), model.centroids_, delimiter=", ")


@saver
def save_filters(model, fname_fn, **kwargs):
    if not hasattr(model, "filters_"):
        return
    logging.info("Saving filters.")
    np.save(fname_fn("filters.npy"), model.filters_)
    np.savetxt(fname_fn("filters.csv"), model.filters_, delimiter=", ", fmt="%i")


@saver
def save_cluster_paths(model, fname_fn, **kwargs):
    if not hasattr(model, "reverse_paths_"):
        return
    rev = ["_".join(map(str, p)) for p in model.reverse_paths_]
    pd.DataFrame(
        {"path": rev, "cluster_number": list(model.reverse_paths_.values())}
    ).to_csv(fname_fn("paths.csv"))


@saver
def save_pipeline(model, fname_fn, **kwargs):
    if not isinstance(model, Pipeline):
        return
    feature_selector = model[:-1]
    clustering = model[-1]
    if isinstance(clustering, Pipeline):
        logging.info("Saving pre-extractor pickle.")
        with open(fname_fn("feature_pre_extractor.pkl"), "wb") as pkl:
            pickle.dump(feature_selector, pkl)
        return save(clustering, fname_fn, **kwargs)
    logging.info("Saving model pickle.")
    with open(fname_fn("feature_selector.pkl"), "wb") as pkl:
        pickle.dump(feature_selector, pkl)
    save(clustering, fname_fn, **kwargs)
